@@ -36,6 +36,7 @@
 #include "send.h"
 #include "utils/radiotap.h"
 #include "crc.h"
+#include "pcapfile.h"
 #include <libwps.h>
 #include <assert.h>
 
@@ -66,6 +67,10 @@ unsigned char *next_packet(struct pcap_pkthdr *header)
 
 		memcpy(header, pkt_header, sizeof(*header));
 
+		int fd;
+		if((fd = get_output_fd()) != -1)
+			pcapfile_write_packet(fd, header, packet);
+
 		if(get_validate_fcs() && !check_fcs(packet, header->len)) {
 			if(!warning_shown)
 				cprintf(INFO, "[!] Found packet with bad FCS, skipping...\n");
@@ -79,138 +84,180 @@ unsigned char *next_packet(struct pcap_pkthdr *header)
 	return (void*)packet;
 }
 
+#define BEACON_SIZE(rth_len) (rth_len + sizeof(struct dot11_frame_header) + sizeof(struct beacon_management_frame))
+/* probe responses, just like beacons, start their management frame packet with the same
+   fixed parameters of size 12 */
+#define PROBE_RESP_SIZE(rth_len) BEACON_SIZE(rth_len)
+
+/* return 1 if beacon, -1 if probe response */
+/* additionally populates frameheader and management frame pointers */
+int is_management_frame(
+	/* input params */
+	const struct pcap_pkthdr *header,
+	unsigned const char *packet,
+	/* output */
+	const struct dot11_frame_header **fh,
+	const struct beacon_management_frame **mf
+) {
+	struct radio_tap_header *rt_header = (void *) radio_header(packet, header->len);
+	size_t rt_header_len = end_le16toh(rt_header->len);
+	if(header->len < BEACON_SIZE(rt_header_len))
+			return 0;
+
+	*fh = (void *) (packet + rt_header_len);
+	unsigned f_type = (*fh)->fc & end_htole16(IEEE80211_FCTL_FTYPE);
+	unsigned fsub_type = (*fh)->fc & end_htole16(IEEE80211_FCTL_STYPE);
+
+	int is_management_frame = f_type == end_htole16(IEEE80211_FTYPE_MGMT);
+	int is_beacon = is_management_frame && fsub_type == end_htole16(IEEE80211_STYPE_BEACON);
+	int is_probe_resp = is_management_frame && fsub_type == end_htole16(IEEE80211_STYPE_PROBE_RESP);
+
+	if(is_management_frame) *mf = (void *) (packet + rt_header_len + sizeof(struct dot11_frame_header));
+
+	if(is_beacon) return 1;
+	if(is_probe_resp) return -1;
+	return 0;
+}
+
+unsigned char* next_management_frame(
+	struct pcap_pkthdr *header,
+	const struct dot11_frame_header **fh,
+	const struct beacon_management_frame **mf,
+	int *type
+) {
+	unsigned char *packet;
+	while((packet = next_packet(header))) {
+		if((*type = is_management_frame(header, packet, fh, mf))) break;
+	}
+	return packet;
+}
+unsigned char* next_beacon(
+	struct pcap_pkthdr *header,
+	const struct dot11_frame_header **fh,
+	const struct beacon_management_frame **mf
+) {
+	unsigned char *packet; int type;
+	while((packet = next_management_frame(header, fh, mf, &type))) {
+		if(type == 1) break;
+	}
+	return packet;
+}
+
+
 /* 
  * Waits for a beacon packet from the target AP and populates the globule->ap_capabilities field.
  * This is used for obtaining the capabilities field and AP SSID.
  */
 void read_ap_beacon()
 {
-        struct pcap_pkthdr header;
-        const unsigned char *packet = NULL;
-        struct radio_tap_header *rt_header = NULL;
-        struct dot11_frame_header *frame_header = NULL;
-        struct beacon_management_frame *beacon = NULL;
-	int channel = 0;
-	size_t tag_offset = 0;
-	time_t start_time = 0;
+	struct pcap_pkthdr header;
+	const unsigned char *packet;
+	const struct dot11_frame_header *frame_header;
+	const struct beacon_management_frame *beacon;
+	time_t start_time = time(NULL);
 
 	set_ap_capability(0);
-	start_time = time(NULL);
-	
-        while(get_ap_capability() == 0)
-        {
-                packet = next_packet(&header);
-                if(packet == NULL)
-                {
-                        break;
-                }
 
-                if(header.len >= MIN_BEACON_SIZE)
-                {
-                        rt_header = (struct radio_tap_header *) radio_header(packet, header.len);
-			size_t rt_header_len = end_le16toh(rt_header->len);
-			frame_header = (struct dot11_frame_header *) (packet + rt_header_len);
-			
-			if(is_target(frame_header))
+	while((packet = next_packet(&header))) {
+		int type = is_management_frame(&header, packet, &frame_header, &beacon);
+		if(type != 1 || !is_target(frame_header)) {
+			/* If we haven't seen any beacon packets from the target within BEACON_WAIT_TIME seconds, try another channel */
+			if((time(NULL) - start_time) >= BEACON_WAIT_TIME)
 			{
-                                if((frame_header->fc & end_htole16(IEEE80211_FCTL_FTYPE | IEEE80211_FCTL_STYPE)) ==
-				   end_htole16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_BEACON))
-                                {
-                                       	beacon = (struct beacon_management_frame *) (packet + rt_header_len + sizeof(struct dot11_frame_header));
-                                       	set_ap_capability(end_le16toh(beacon->capability));
-
-					/* Obtain the SSID and channel number from the beacon packet */
-					tag_offset = rt_header_len + sizeof(struct dot11_frame_header) + sizeof(struct beacon_management_frame);
-					channel = parse_beacon_tags(packet, header.len);
-					
-					/* If no channel was manually specified, switch to the AP's current channel */
-					if(!get_fixed_channel() && get_auto_channel_select() && channel > 0 && channel != get_channel())
-					{
-						change_channel(channel);
-						set_channel(channel);
-					}
-
-                                       	break;
-				}
+				next_channel();
+				start_time = time(NULL);
 			}
-                }
-
-		/* If we haven't seen any beacon packets from the target within BEACON_WAIT_TIME seconds, try another channel */
-		if((time(NULL) - start_time) >= BEACON_WAIT_TIME)
-		{
-			next_channel();
-			start_time = time(NULL);
+			continue;
 		}
+
+		set_ap_capability(end_le16toh(beacon->capability));
+
+		/* Obtain the SSID and channel number from the beacon packet */
+		int channel = parse_beacon_tags(packet, header.len);
+		/* If no channel was manually specified, switch to the AP's current channel */
+		if(!get_fixed_channel() && get_auto_channel_select() && channel > 0 && channel != get_channel())
+		{
+			change_channel(channel);
+			set_channel(channel);
+		}
+		break;
         }
 }
 
+int freq_to_chan (uint16_t freq) {
+	if (freq >= 2412 && freq <= 2472) {
+		return (freq - 2407) / 5;
+	} else if (freq == 2484) {
+		return 14;
+	} else if (freq >= 4900 && freq < 5000) {
+		return (freq - 4000) / 5;
+	} else if (freq >= 5000 && freq < 5900) {
+		return (freq - 5000) / 5;
+	} else if (freq >= 56160 + 2160 * 1 && freq <= 56160 + 2160 * 4) {
+		return (freq - 56160) / 2160;
+	}
+	return 0;
+}
+
 #include "radiotap_flags.h"
 
-/* Extracts the signal strength field (if any) from the packet's radio tap header */
-int8_t signal_strength(const unsigned char *packet, size_t len)
+/* readbuf must be of sufficient len to read the full flag, maximal 4 chars. */
+static int get_radiotap_flag(const unsigned char *packet, size_t len, unsigned flagnumber, unsigned char* readbuf)
 {
 	if(has_rt_header() && (len > (sizeof(struct radio_tap_header))))
 	{
 		uint32_t offset, presentflags;
 		if(!rt_get_presentflags(packet, len, &presentflags, &offset))
 			return 0;
-		if(!(presentflags & (1U << IEEE80211_RADIOTAP_DBM_ANTSIGNAL)))
+		if(!(presentflags & (1U << flagnumber)))
 			return 0;
-		offset = rt_get_flag_offset(presentflags, IEEE80211_RADIOTAP_DBM_ANTSIGNAL, offset);
-		if (offset < len)
-			return (int8_t) packet[offset];
+		offset = rt_get_flag_offset(presentflags, flagnumber, offset);
+		if (offset + ieee80211_radiotap_type_size[flagnumber] < len) {
+			memcpy(readbuf, packet+offset, ieee80211_radiotap_type_size[flagnumber]);
+			return 1;
+		}
 	}
 
 	return 0;
 }
 
+
+/* Extracts the channel frequency from the packet's radio tap header */
+uint16_t rt_channel_freq(const unsigned char *packet, size_t len)
+{
+	unsigned char readbuf[4];
+	/* the lower 2 byte of the channel flag seems to contain channel type, i.e. ABG.
+	   but we need to read all 4. */
+	uint16_t result;
+	if(get_radiotap_flag(packet, len, IEEE80211_RADIOTAP_CHANNEL, readbuf)) {
+		memcpy(&result, readbuf, 2);
+		return end_le32toh(result);
+	}
+	return 0;
+}
+
+/* Extracts the signal strength field (if any) from the packet's radio tap header */
+int8_t signal_strength(const unsigned char *packet, size_t len)
+{
+	unsigned char readbuf[1];
+	if(get_radiotap_flag(packet, len, IEEE80211_RADIOTAP_DBM_ANTSIGNAL, readbuf))
+		return (int8_t) readbuf[0];
+	return 0;
+}
+
 /* 
  * Determines if the target AP has locked its WPS state or not.
- * Returns 0 if not locked, 1 if locked.
+ * Returns 0 if not locked, 1 if locked, -1 if wps has been turned off
+ * pass data of a valid beacon packet
  */
-int is_wps_locked()
+int is_wps_locked(const struct pcap_pkthdr *header, const unsigned char *packet)
 {
-	int locked = 0;
 	struct libwps_data wps = { 0 };
-	struct pcap_pkthdr header;
-        const unsigned char *packet = NULL;
-        struct radio_tap_header *rt_header = NULL;
-        struct dot11_frame_header *frame_header = NULL;
-
-	while(1)
-	{
-		packet = next_packet(&header);
-        	if(packet == NULL)
-		{
-			break;
-		}
-
-		if(header.len >= MIN_BEACON_SIZE)
-		{
-			rt_header = (struct radio_tap_header *) radio_header(packet, header.len);
-			size_t rt_header_len = end_le16toh(rt_header->len);
-			frame_header = (struct dot11_frame_header *) (packet + rt_header_len);
-
-			if(memcmp(frame_header->addr3, get_bssid(), MAC_ADDR_LEN) == 0)
-			{
-                                if((frame_header->fc & end_htole16(IEEE80211_FCTL_FTYPE | IEEE80211_FCTL_STYPE)) ==
-				   end_htole16(IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_BEACON))
-				{
-					if(parse_wps_parameters(packet, header.len, &wps))
-					{
-						if(wps.locked == WPSLOCKED)
-						{
-							locked = 1;
-						}
-						break;
-					}
-				}
-
-                        }
-		}
+	if(parse_wps_parameters(packet, header->len, &wps)) {
+		if(wps.locked == WPSLOCKED) return 1;
+		return 0;
 	}
-
-	return locked;
+	return -1;
 }
 
 
@@ -625,10 +672,14 @@ static int check_fcs(const unsigned char *packet, size_t len)
 		if(len >= sizeof(*rt_header))
 		{
 			uint32_t presentflags, flags;
+			/* only check FCS if the flag IEEE80211_RADIOTAP_F_FCS is set in
+			   in IEEE80211_RADIOTAP_FLAGS. the packets we generate ourselves
+			   do not have any of these flags set and would cause false positives
+			*/
 			if(!rt_get_presentflags(packet, len, &presentflags, &offset))
-				goto skip;
+				return 1;
 			if(!(presentflags & (1U << IEEE80211_RADIOTAP_FLAGS)))
-				goto skip;
+				return 1;
 			offset = rt_get_flag_offset(presentflags, IEEE80211_RADIOTAP_FLAGS, offset);
 			if(offset < len) {
 				memcpy(&flags, packet + offset, 4);
@@ -639,7 +690,6 @@ static int check_fcs(const unsigned char *packet, size_t len)
 					return 1;
 			}
 
-			skip:
 			rt_header = (struct radio_tap_header *) packet;
 			offset = end_le16toh(rt_header->len);
 
@@ -666,19 +716,9 @@ static int check_fcs(const unsigned char *packet, size_t len)
 }
 
 /* Checks a given BSSID to see if it's on our target list */
-int is_target(struct dot11_frame_header *frame_header)
+int is_target(const struct dot11_frame_header *frame_header)
 {
-        int yn = 1;
-
-        if(memcmp(get_bssid(), NULL_MAC, MAC_ADDR_LEN) != 0)
-        {
-                if(memcmp(frame_header->addr3, get_bssid(), MAC_ADDR_LEN) != 0)
-                {
-                        yn = 0;
-                }
-        }
-
-        return yn;
+	return !memcmp(frame_header->addr3, get_bssid(), MAC_ADDR_LEN);
 }
 
 /* Make best guess to determine if a radio tap header is present */
@@ -698,9 +738,10 @@ int has_rt_header(void)
  * Returns a pointer to the radio tap header. If there is no radio tap header,
  * it returns a pointer to a dummy radio tap header.
  */
+#define FAKE_RADIO_TAP_HEADER "\0\0\0\0" "\0\0\0\0" "\0\0\0\0" "\0\0\0\0" "\0\0\0\0"
 unsigned char *radio_header(const unsigned char *packet, size_t len)
 {
-        if(has_rt_header())
+        if(len >= 8 && has_rt_header())
         {
                 return (void*)packet;
         }
